#include-once
#include <File.au3>
;#include <DebugWrite.au3>

Const $DW_CONTINUE=0
Const $DW_BREAK=1
Const $DW_BREAK_ALL=2

;-----------------------------------------------------------------------------
; a few internal functions

Const $_DW_ARRAY_BOUND=64
Global $g_dwPattern,$g_dwFiles[1],$g_dwRecurse,$g_dwInvert,$g_dwFileCB,$g_dwDirCB,$g_dwLevel

; Internal default callback for files
; $name is the full filename (i.e. including path)
; $isdir is <>0 if $name is a directory
; $level is the directory level (good if called recursively)
;
; returns $DW_CONTINUE to continue the walk
; or $DW_BREAK to stop current directory walk
; or $DW_BREAK_ALL to stop walking completely
Func _InternalFileCB($name,$isdir,$level)
	$g_dwFiles[0]+=1
	If $g_dwFiles[0]=UBound($g_dwFiles) Then ReDim $g_dwFiles[UBound($g_dwFiles)+$_DW_ARRAY_BOUND]
	$g_dwFiles[$g_dwFiles[0]]=$name
	If $isdir Then
		;DebugWrite("Dir:  %s (%i)",$name,$level)
	Else
		;DebugWrite("File: %s (%i)",$name,$level)
	EndIf
	Return $DW_CONTINUE
EndFunc

; Internal default callback for directories
; $name is the full directory name (i.e. including path)
; $level is the directory level
;
; return values see _InternalFileCB()
Func _InternalDirCB($name,$level)
  If $level>0 Then
	  ;DebugWrite("Entry: %s (%i)",$name,$level)
	Else
	  ;DebugWrite("Exit : %s (%i)",$name,$level)
	EndIf
	Return $DW_CONTINUE
EndFunc

; Internal workhorse, can call itself recursively
Func _DirWalk_recurse($path)
	Local $i=0,$hfs,$full,$files[$_DW_ARRAY_BOUND],$c,$d
	If StringRight($path,1)<>"\" Then $path&="\"
	$hfs=FileFindFirstFile($path&"*")
	If $hfs=-1 Then Return
	While 1
		$i+=1
		If $i=UBound($files) Then ReDim $files[UBound($files)+$_DW_ARRAY_BOUND]
		$files[$i]=FileFindNextFile($hfs)
		If @error Then ExitLoop
	WEnd
	FileClose($hfs)
	$files[0]=$i-1
	For $i=1 To $files[0]
		$full=$path&$files[$i]
		$d=_IsDirectory($full)
		If BitXor($g_dwInvert,StringRegExp($files[$i],$g_dwPattern)) Then
			$c=Call($g_dwFileCB,$full,$d,$g_dwLevel)
			If $c<>$DW_CONTINUE Then ExitLoop
		EndIf
		If $d And $g_dwRecurse Then
			$g_dwLevel+=1
			$c=$DW_CONTINUE
			If Call($g_dwDirCB,$full,$g_dwLevel)=$DW_CONTINUE Then $c=_DirWalk_recurse($full)
			Call($g_dwDirCB,$full,-$g_dwLevel)
			$g_dwLevel-=1
			If $c=$DW_BREAK_ALL Then ExitLoop
		EndIf
	Next
	Return $c
EndFunc

;----------------------------------------------------------------------------
; External functions

; _IsDirectory($s)
; Helper function: returns a value>0 if $s is a directory
Func _IsDirectory($s)
	Return StringInStr(FileGetAttrib($s),"D")
EndFunc

; _DirWalk($path=".\",$pattern="*",$flags=0,$filecb="_InternalFileCB",$dircb="_InternalDirCB")
; Walks a directory and can return the files/directories found
;
; $path is the path to look in; this can be c:\ or d:\Windows
; $pattern is the pattern to look for; this can either be an OS-type wildcard
; (i.e. *.do?) or a regular expression (see $flags)
;
; The flags are bit-coded values and can be freely combined:
; if $flags=1 then $path is walked down recursively
; if $flags=2 then $pattern is a regular expression (don't forget ^...$ if
; you want the pattern anchored)
; if $flags=4 then the result of the pattern match is inverted (i.e. w*.exe
; will return all files *except* w*.exe files)
;
; if $filecb is defined it must be the name of a function that is used as a
; callback for all files and directories found; see _InternalFileCB()
; if $dircb is defined it must be the name of a function that is used as a
; callback before/after recursing down/up a directory; see _InternalDirCB()
; If either callback is not defined the internal default is used: _InternalFileCB()
; stores all files and directories in an array which is returned by this function
; _InternalDirCB() does nothing.
Func _DirWalk($path=".\",$pattern="*",$flags=0,$filecb="_InternalFileCB",$dircb="_InternalDirCB")
	$g_dwRecurse=BitAnd($flags,1)
	$g_dwInvert=0
	If BitAnd($flags,4) Then $g_dwInvert=1 ; may look silly but don't change
	$g_dwFiles[0]=0
	$g_dwFileCB=$filecb
	$g_dwDirCB=$dircb
	If BitAnd($flags,2) Then
		$g_dwPattern=$pattern
	Else	
		$pattern=StringReplace($pattern,".","\.")
		$pattern=StringReplace($pattern,"?",".")
		$g_dwPattern=StringReplace($pattern,"*",".*")
		$g_dwPattern="^(?i)"&$g_dwPattern&"$"
	EndIf
	$g_dwLevel=1
	_DirWalk_recurse($path)
	Return $g_dwFiles
EndFunc

; _DirSplit(Byref $pattern,Byref $path)
; Helper function: initialises a wildcard pattern/path combination such that
; it can used for _DirWalk(): upon calling, $pattern holds a directory
; and/or pattern. This is split, so the upon return $pattern has just the
; pattern part and $path... well, the path:-). In other words, $path is
; overwritten by this code. 
; This function does NOT support regular expressions, only wildcards!
Func _DirSplit(Byref $pattern,Byref $path)
	Local $d1,$d2,$d3
	If $pattern<>"*" Then
		If _IsDirectory($pattern) Then
			$path=$pattern
			$pattern="*"
		Else
			_PathSplit($pattern,$path,$d1,$d2,$d3)
			;DebugWrite("After split: !%s!%s!%s!%s!",$path,$d1,$d2,$d3)
			$path&=$d1
			$pattern=$d2&$d3
			If $pattern="" Then $pattern="*"
		EndIf	
	Else
		$path=".\"
	EndIf
EndFunc
